/*  cpi
 Copyright (C) Tomas Kaiser

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include "cpi.h"

// #define DEBUG	1

static char *progname = NULL;
static char usagemsg[ 256 ] = "";

static char in_filename[ 256 ] = "";
static char out_filename[ 256 ] = "";
static FILE *fr = NULL;
static FILE *fw = NULL;

static struct ms_cpi_file_header ffh;
static struct font_info_header fih;
static struct cp_entry_header ceh;
static struct cp_info_header cih;
static struct cp_fontdata cfd;

/*-------------------------------------------------------------------------*/

void runerror( int status, char *format, ... )
{
  char buffer[ 512 ];
  va_list arglist;

  va_start( arglist, format );
  vsprintf( buffer, format, arglist );
  fprintf( stderr, "%s: %s\n", progname, buffer );
  va_end( arglist );

  exit( status );
}



void usage( int status )
{
  fprintf( stderr, "%s\n", usagemsg );
  exit( status );
}



size_t _fread( void *ptr, size_t size, size_t nmemb, FILE *stream)
{
  if ( fread( ptr, size, nmemb, stream ) != nmemb )
    runerror( -1, "Error readding from file %s", in_filename );

  return nmemb;
}



int _fprintf(FILE *stream, const char *format, ...)
{
  char buffer[ 1024 ];
  va_list arglist;
  int rval;

  va_start( arglist, format );
  vsprintf( buffer, format, arglist );
  if ( ( rval = fprintf( stream, buffer ) ) < 0 )
    runerror( -1, "Error writing to file %s", out_filename );
  va_end( arglist );

  return rval;
}

/*-------------------------------------------------------------------------*/

#if DEBUG

void trace( char *format, ... )
{
  char buffer[ 1024 ];
  va_list arglist;

  va_start( arglist, format );
  vsprintf( buffer, format, arglist );
  fprintf( stderr, "%s: %s\n", progname, buffer );
  fflush( stderr );
  va_end( arglist );
}

#else

#define  trace( format, ... ) while( 0 ) { } 

#endif

/*-------------------------------------------------------------------------*/

void read_reserved( const int nr, const int shift )
{
  int i;
  u08 b;

  for ( i = 0; i < nr; i++ )
  {
    _fread( &b, sizeof( u08 ), 1, fr );
    trace( "%*sReserved %02u: 0x%02X", shift, "", i+1, b );
  }
}


void read_ms_cpi_file_header( void )
{
  trace( "[ font file header ]" );
  _fread( &ffh.file_tag, sizeof( u08 ), 1, fr );
  trace( "File tag: 0x%02X", ffh.file_tag );
  _fread( ffh.font_string, sizeof( char ), 7, fr );
  ffh.font_string[ 7 ] = '\0';
  trace( "Font string: '%s'", ffh.font_string );
  read_reserved( 8, 0 );
  _fread( &ffh.num_pointers, sizeof( u16 ), 1, fr );
  trace( "Number of pointers: %u", ffh.num_pointers );
  _fread( &ffh.p_type, sizeof( u08 ), 1, fr );
  trace( "Type of pointer: %u", ffh.p_type );
  _fread( &ffh.offset, sizeof( u32 ), 1, fr );
  trace( "Offset to info: 0x%04X(%u)", ffh.offset, ffh.offset );
}



void read_font_info_header( void )
{
  trace( "  [ font info header ]" );
  _fread( &fih.nb_codepages, sizeof( u16 ), 1, fr );
  trace( "  Number of codepages: %u", fih.nb_codepages );
}

void read_cp_entry_header( void )
{
  trace( "    [ cp entry header ]" );
  _fread( &ceh.cp_header_size, sizeof( u16 ), 1, fr );
  trace( "    CP header size: 0x%02X(%u)",
    ceh.cp_header_size, ceh.cp_header_size );
  _fread( &ceh.next_header_offset, sizeof( u32 ), 1, fr );
  trace( "    Next header offset: 0x%04X(%u)", 
    ceh.next_header_offset, ceh.next_header_offset );
  _fread( &ceh.device_type, sizeof( u16 ), 1, fr );
  trace( "    Device type: %u", ceh.device_type );
  _fread( ceh.device_name, sizeof( char ), 8, fr );
  ceh.device_name[ 8 ] = '\0';
  trace( "    Device name (ID): %s", ceh.device_name );
  _fread( &ceh.cp_number, sizeof( u16 ), 1, fr );
  trace( "    Code page: %u", ceh.cp_number );
  read_reserved( 6, 4 );
  _fread( &ceh.font_offset, sizeof( u32 ), 1, fr );
  trace( "    Font info offset: 0x%04X(%u)", ceh.font_offset, ceh.font_offset );
}

void read_cp_info_header( void )
{
  trace( "      [ font data header ]" );
  read_reserved( 2, 6 );
  _fread( &cih.num_fonts, sizeof( u16 ), 1, fr );
  trace( "      Number of fonts: %u", cih.num_fonts );
  _fread( &cih.bitmap_size, sizeof( u16 ), 1, fr );
  trace( "      Font data length: 0x%02x(%u)",
    cih.bitmap_size, cih.bitmap_size );
}

void read_cp_fontdata( void )
{
  trace( "        [ font data ]" );
  _fread( &cfd.charbox_rows, sizeof( u08 ), 1, fr );
  trace( "        Rows in character box: %u", cfd.charbox_rows );
  _fread( &cfd.charbox_cols, sizeof( u08 ), 1, fr );
  trace( "        Columns in character box: %u", cfd.charbox_cols );
  _fread( &cfd.aspect_ratio, sizeof( u16 ), 1, fr );
  trace( "        Aspect ratio (not used): %u", cfd.aspect_ratio );
  _fread( &cfd.num_chars, sizeof( u16 ), 1, fr );
  trace( "        Number of characters in font: %u", cfd.num_chars );
}


void process_fonts( const u16 n, const u08 r, const u08 c )
{
  u08 b;
  u16 i;
  u08 j;
  u08 k;

  ( void )k;
  sprintf( out_filename, "vga%u-%ux%u.bdf", ceh.cp_number, c, r );
  if ( ( fw = fopen( out_filename, "w" ) ) == NULL )
    runerror( -1, "Failed opening %s for writing.", out_filename );
 
  _fprintf( fw, "STARTFONT 2.1\n" );
  _fprintf( fw, "FONT vga%u-%ux%u\n", ceh.cp_number, c, r );
  if ( r == 8 )
  {
    _fprintf( fw, "SIZE 8 75 75\n" );
    _fprintf( fw, "FONTBOUNDINGBOX 8 8 0 -2\n" );
  }
  else if ( r == 14 )
  {
    _fprintf( fw, "SIZE 14 75 75\n" );
    _fprintf( fw, "FONTBOUNDINGBOX 8 14 0 -3\n" );
  }
  else if ( r == 16 )
  {
    _fprintf( fw, "SIZE 16 75 75\n" );
    _fprintf( fw, "FONTBOUNDINGBOX 8 16 0 -4\n" );
  }
   _fprintf( fw, "STARTPROPERTIES 3\n" );
   _fprintf( fw, "FONT_DESCENT 4\n" );
   _fprintf( fw, "FONT_ASCENT 12\n" );
   _fprintf( fw, "DEFAULT_CHAR 0\n" );
   _fprintf( fw, "ENDPROPERTIES\n" );
   _fprintf( fw, "CHARS %u\n", n );
   
  
  for( i = 0; i < n; i++ )
  {
    _fprintf( fw, "STARTCHAR C%04x\n", i );
    _fprintf( fw, "ENCODING %u\n", i );
    _fprintf( fw, "SWIDTH 480 0\n" );
    _fprintf( fw, "DWIDTH 8 0\n" );
    if ( r == 8 )
      _fprintf( fw, "BBX 8 8 0 -2\n" );
    else if ( r == 14 )
      _fprintf( fw, "BBX 8 14 0 -3\n" );
    else if ( r == 16 )
      _fprintf( fw, "BBX 8 16 0 -4\n" );
    _fprintf( fw, "BITMAP\n" );
	    
    for( j = 0; j < r; j++ )
    {
      _fread( &b, sizeof( u08 ), 1, fr );
      _fprintf( fw, "%02X\n", b );
    }

    _fprintf( fw, "ENDCHAR\n" );
  }

  _fprintf( fw, "ENDFONT\n" );
  fclose( fw );
}

/*-------------------------------------------------------------------------*/

int main( int argc, char *argv[] )
{
  u16 i, j;
	
  progname = strrchr( argv[ 0 ], '/' );
  progname = ( progname != NULL ) ? ++progname : argv[ 0 ];
  sprintf( usagemsg, "Usage: %s <cpifile>", progname );

  if ( argc != 2 )
    usage( -1 );
  
  strcpy( in_filename, argv[ 1 ] );
  if ( ( fr = fopen( in_filename, "r" ) ) == NULL )
    runerror( -1, "Failed opening %s for reading.", in_filename );
 
  read_ms_cpi_file_header();
  read_font_info_header();
  for ( i = 0; i < fih.nb_codepages; i++ )
  {
    read_cp_entry_header();
    read_cp_info_header();
    for ( j = 0; j < cih.num_fonts; j++ )
    {
      read_cp_fontdata();
      process_fonts( cfd.num_chars, cfd.charbox_rows, cfd.charbox_cols );
    }
  }

  fclose( fr );
  return 0;
}

